from causaldag.classes import DAG
import numpy as np
from causaldag.utils import core_utils
from typing import Callable, Dict, Hashable, Optional
from tqdm import tqdm


class CamDAG(DAG):
    def __init__(self, nodes, arcs):
        super().__init__(set(nodes), arcs)
        self.mean_functions = dict()
        self.node_noises = dict()
        self._node_list = list(nodes)
        self._node2ix = core_utils.ix_map_from_list(self._node_list)

    def set_mean_function(self, node, mean_function: Callable[[np.ndarray, list], np.ndarray]):
        self.mean_functions[node] = mean_function

    def set_noise(self, node, noise_function: Callable[[], np.ndarray]):
        self.node_noises[node] = noise_function

    def sample(self, nsamples: int = 1, progress=False) -> np.array:
        samples = np.zeros((nsamples, len(self._nodes)))
        t = self.topological_sort()
        t = t if not progress else tqdm(t)

        for node in t:
            parents = list(self._parents[node])
            if len(parents) > 0:
                parent_ixs = [self._node2ix[p] for p in parents]
                means = self.mean_functions[node](samples[:, parent_ixs], parents)
            else:
                means = np.zeros(nsamples)
            noise = self.node_noises[node](size=nsamples)
            samples[:, self._node2ix[node]] = means + noise

        return samples

    def conditional_mean(self, cond_values: np.ndarray, cond_nodes: list, check_valid=False) -> np.ndarray:
        assert len(cond_nodes) == cond_values.shape[1]
        if check_valid:
            parents_of_cond = set.union(*(self.parents_of(node) for node in cond_nodes))
            assert parents_of_cond <= set(cond_nodes)
        t = self.topological_sort()
        remaining_nodes = [node for node in t if node not in set(cond_nodes)]

        nsamples = cond_values.shape[0]
        conditional_means = np.zeros((nsamples, self.nnodes))
        conditional_means[:, [self._node2ix[node] for node in cond_nodes]] = cond_values

        for node in remaining_nodes:
            parents = list(self._parents[node])
            parent_ixs = [self._node2ix[p] for p in parents]
            means = self.mean_functions[node](conditional_means[:, parent_ixs], parents)
            conditional_means[:, self._node2ix[node]] = means

        return conditional_means








from sklearn.utils.random import sample_without_replacement
import numpy as np
from joblib import Parallel, delayed
from frozendict import frozendict
from collections import Counter
import ipdb
from sklearn.utils import safe_mask
import random


def bootstrap_generator(n_bootstrap_iterations, sample_fraction, nsamples, random_state=None):
    """Generates bootstrap samples from dataset."""
    if random_state is not None:
        np.random.seed(random_state)
        random.seed(random_state)
    n_subsamples = np.floor(sample_fraction * nsamples).astype(int)
    for _ in range(n_bootstrap_iterations):
        subsample = sample_without_replacement(nsamples, n_subsamples)
        yield subsample


def stability_selection(
        method
):
    def wrapped_method(
            *args,
            parameter_grid: list,
            n_jobs: int = 1,
            n_bootstrap_iterations: int = 50,
            random_state: int = 0,
            sample_fraction: float = 0.7,
            verbose: bool = False,
            bootstrap_threshold: float = 0.5,
            **kwargs
    ):
        parameters2results = dict()

        for parameters in parameter_grid:
            arg2subsamples_list = []
            for X in args:
                nsamples = X.shape[0]
                bootstrap_samples = list(bootstrap_generator(n_bootstrap_iterations, sample_fraction, nsamples, random_state=random_state))
                arg2subsamples_list.append(bootstrap_samples)

            bootstrap_results = Parallel(n_jobs, verbose=verbose)(
                delayed(method)(
                    *(arg[safe_mask(arg, subsample), :] for arg, subsample in zip(args, arg2subsamples)),
                    **kwargs,
                    **parameters
                ) for arg2subsamples in zip(*arg2subsamples_list)
            )

            parameters2results[frozendict(parameters)] = bootstrap_results

        stable_results = set()
        for param, results in parameters2results.items():
            counter = Counter()
            for result in results:
                counter.update(result[0])
            for item, count in counter.items():
                if count >= bootstrap_threshold*n_bootstrap_iterations:
                    stable_results.add(item)

        return stable_results, parameters2results

    return wrapped_method



from ..rand.graphs import directed_erdos, alter_weights, rand_weights, unif_away_original, unif_away_zero
import numpy as np
import random


def create_synthetic_difference(
        nnodes=10,
        nsamples=1000,
        num_altered=1,
        num_removed=1,
        num_added=1,
        exp_nbrs=2,
        seed=8181818,
        return_graphs=False,
        rand_weight_fn=unif_away_zero,
        rand_change_fn=unif_away_original
):
    """
    Create two synthetic datasets from two related causal graphs.

    nnodes:
        Number of nodes in each graph.
    nsamples:
        Number of samples from each graph.
    num_altered:
        Number of altered
    """
    random.seed(seed)
    np.random.seed(seed)
    d1 = directed_erdos(nnodes, exp_nbrs=min(exp_nbrs, nnodes-1))
    g1 = rand_weights(d1, rand_weight_fn=rand_weight_fn)
    g2 = alter_weights(
        g1,
        num_altered=num_altered,
        num_removed=num_removed,
        num_added=num_added,
        rand_weight_fn=rand_weight_fn,
        rand_change_fn=rand_change_fn
    )
    X1 = g1.sample(nsamples)
    X2 = g2.sample(nsamples)
    difference = set(zip(*np.where(g1.to_amat() != g2.to_amat())))
    difference_ug = set(zip(*np.where(~np.isclose(g1.precision, g2.precision))))
    difference_ug = {frozenset({i, j}) for i, j in difference_ug if i != j}

    if return_graphs:
        return X1, X2, difference, difference_ug, g1, g2
    return X1, X2, difference


